/*
	ZillionServer.m
	Tue Aug 27 17:02:16 CEST 2002
 */

#import	"ZillionServer.h"
#import	"NSProcessInfo+getopt.h"

/*
	Outline of the protocol:
	The server starts up and offers up his services under the name ZILLION_SERVER.
	A client can register via DO with the server to offer calculation power.
	(-registerClient:)
		- the client has properties:
			- load: how many jobs are submitted in parallel
			- memory: how much memory is available
			- ... to be extended

	A Job is then initiated as follows:
	- The server receives a -dispatchJob: method. This takes a NSData object representing
		a bundle which encodes the job. A job Id is returned.

	- The job is then initiated by choosing from the clients and sending a -loadJob:
		method forwarding the bundle from the server
	- The server then asks the job for a seedEnumerator and for a seed
	- This seed is transferred to the client via -dispatchUnitFromSeed:forJobId:
	- which in turn detaches a new thread and asks the job to -startJobWithClient:andSeed:
	- results are returned 

 */

#define	PROXYSERVER_FOR_JOBID(jobId) (!_respawnStatus? self: [self serverForJobId:jobId])

#define	RESCHEDULE()	[self reschedule]

#define	SEALJOBWITHID(jobId) do { \
	[[NSRunLoop currentRunLoop] performSelector:@selector(sealJobWithId:) \
		target:self \
		argument:[jobId copy] order:1 \
		modes:[NSArray arrayWithObject:NSDefaultRunLoopMode]]; \
	NSLog(@"scheduled sealing of job:%@", jobId);\
	} while (0)

#define	JobSeedStatus(j) \
	(!![[self jobSpecificationForJobId:j] objectForKey:@"seedsAreExhausted"])
#define	SetJobSeedsExhausted(j) \
	[[self jobSpecificationForJobId:j] setObject:@"YES" forKey:@"seedsAreExhausted"]


@implementation ZillionServer

/*
	The idea for respawning a new server is that classes cannot
	be unloaded in objc currently due to reentrance problems which would occur otherwise.

	The new server is started and 
 */

- (NSString *)DOname { return ZILLION_SERVER_NAME; }
- (NSArray *)respawningArgumentsWithId:(NSString *)doNameId {
	return [NSArray arrayWithObjects:@"--serverName", [self respawningDOnameWithId:doNameId],
		@"--doRespawn=NO", nil];
}
- (Protocol *)respawningProtocol { return @protocol(ZillionServerProtocol); }

- initWithRespawnStatus:(BOOL)respawnStatus
{
	if (!(self = [super init])) return self;

	clients = [[NSMutableDictionary alloc] init];
	_respawnStatus = respawnStatus;
	proxyServers = [[NSMutableDictionary alloc] init];

	outcomes = [[NSMutableDictionary alloc] init];
	dispatchedSeeds = [[NSMutableDictionary alloc] init];

	return self;
}
- init
{
	return [self initWithRespawnStatus:NO];
}

- (void)dealloc
{
	NSLog(@"about to dealloc ZillionServer object");
	[dispatchedSeeds release];
	[outcomes release];

	[clients release];
	[proxyServers release];
	[super dealloc];
}

- (void)reschedule
{
	if (_reschedulingIsMarked) return;
	_reschedulingIsMarked = YES;
	[[NSRunLoop currentRunLoop] performSelector:@selector(scheduleJobs:)
		target:self argument:nil order:0 modes:[NSArray arrayWithObject:NSDefaultRunLoopMode]];
}


- (void)clientConnectionDidDie:(NSNotification *)notification
{
	// presume this is the client <t>
	[self deregisterClient:[[notification object] rootObject]];
}

- (void)registerClient:(id <ZillionClientProtocol>)someClient
{
	NSDictionary	*newClient = [NSDictionary dictionaryWithObjectsAndKeys:
		someClient, @"object", [someClient attributes], @"attributes", nil];

	[clients setObject:newClient forKey:[someClient clientId]];

	// <!> see doc for possible drawbacks; we might not notice invalidation
	// unless we ping the client <i>
	[[NSNotificationCenter defaultCenter] addObserver:self
		selector:@selector(clientConnectionDidDie:)
		name:NSConnectionDidDieNotification
		object:[(NSDistantObject *)someClient connectionForProxy]];

	NSLog(@"Registered client:%@", someClient);
	RESCHEDULE();
}

- (void)deregisterClient:(id <ZillionClientProtocol>)someClient
{
	[clients removeObjectForKey:[someClient clientId]];
	RESCHEDULE();
}

- (id <ZillionServerProtocol>)serverForJobId:(NSString *)jobId {
	return [proxyServers objectForKey:jobId];
}


- (void)respawnedConnectionDidDie:(NSNotification *)notification
{
	NSLog(@"Some proxy went away");
}

- (NSString *)dispatchJob:(NSData *)bundle
{
	id <ZillionServerProtocol>	proxyServer = self;
	NSString	*jobId;

	//	we respawn a server to loadd the job 
	if (_respawnStatus) {
		proxyServer = [self respawnedObjectWithId:
			[[NSNumber numberWithInt:proxyServerIdCounter] stringValue]
			doObserveConnection:YES];
		[proxyServer setMasterServer:self];
		proxyServerIdCounter++;
	}

	jobId = [proxyServer loadAndRegisterJob:bundle];

	if (_respawnStatus) {
		[proxyServers setObject:proxyServer forKey:jobId];
	}
	RESCHEDULE();
	return jobId;
}

- (NSString *)description { NSLog(@"description forthcoming");
	return [super description];
}

//	called in the proxyServer

- (void)sealJobWithId:(NSString *)jobId
{
	NSLog(@"Welcome to the finished job [%d]:%@", _respawnStatus, jobId);
	if (_respawnStatus) {
		id <ZillionServerProtocol>	proxy = PROXYSERVER_FOR_JOBID(jobId);
		[proxy sealJobWithId:jobId];
		NSLog(@"did seal job in proxy for job:%@, now terminating", jobId);
		[proxy terminate];
		[proxyServers removeObjectForKey:jobId];
	} else {
		[[[self jobInstance:jobId] class] sealJobOutcomesUsingServer:[self masterServer]];
		NSLog(@"Sealing job:%@", jobId);
	}
}

- (BOOL)serverToGoDown { return _serverToGoDown; }
- (void)terminate { 
	NSLog(@"Server is about to terminate");
	if (_respawnStatus) {
		NSEnumerator				*proxies = [proxyServers objectEnumerator];
		id <ZillionServerProtocol>	proxyServer;
		while ( (proxyServer = [proxies nextObject]) )
			[proxyServer terminate];
	}
	_serverToGoDown = YES;
}

- (void)registerJob:(NSData *)jobBundle withId:(NSString *)jobId
{
	NSMutableDictionary	*jobSpec = [self jobSpecificationForJobId:jobId];

	NSLog(@"Registering %@.", jobId);
	[jobSpec setObject:jobBundle forKey:@"packedBundle"];
	[jobSpec setObject:[NSNumber numberWithDouble:0] forKey:@"timePenalty"];
}
- (NSData *)packedJobBundleForJobId:(NSString *)jobId
{
	return [[self jobSpecificationForJobId:jobId] objectForKey:@"packedBundle"];
}
- (NSString *)loadAndRegisterJob:(NSData *)packedBundle
{
	NSBundle	*jobBundle = [self bundleForPackedBundle:packedBundle];
	Class		jobClass = [jobBundle principalClass];

	NSLog(@"...about to register %lx.", [self masterServer]);
	[[self masterServer] registerJob:packedBundle withId:[jobClass jobId]];
	[self instantiateJobFromBundle:jobBundle];
	NSLog(@"...loaded the code for job %@.", [jobClass jobId]);
	return [jobClass jobId];
}
/*
	this returns a unique jobInstance
 */
- (id <ZillionJobProtocol>)jobInstance:(NSString *)jobId
{
	NSMutableDictionary	*jobSpec = [self jobSpecificationForJobId:jobId];
	id <ZillionJobProtocol>	job = [jobSpec objectForKey:@"instance"];

	/*	we must be aware that any instances might have gone before
		sind on reloading the proxy server is done away with to free
		the class hierarchy of ObjC
	 */
	if (!job) {
		job = [self instantiateJobWithId:jobId];
		[jobSpec setObject:job forKey:@"instance"];
	}
	return job;
}

- (NSData *)nextSeedIdForJobId:(NSString *)jobId
{
	// trigger job instantiation
	id <ZillionJobProtocol>	job = [self jobInstance:jobId];
	NSMutableDictionary		*jobSpec = [self jobSpecificationForJobId:jobId];
	NSEnumerator			*seedEnum = [jobSpec objectForKey:@"enumerator"];

	if (!seedEnum) {
		NSLog(@"new seed enum");
		seedEnum = [job seedEnumeratorWithSeed:nil];
		[jobSpec setObject:seedEnum forKey:@"enumerator"];
	}
	return [[job class] dataRepresentationOfSeed:[seedEnum nextObject]];
}

#define	JOBPenalty(j) \
	[[[self jobSpecificationForJobId:j] objectForKey:@"timePenalty"] doubleValue]
#define	JOBNextSeed(j) \
	[[[self jobSpecificationForJobId:j] objectForKey:@"seedEnumerator"] nextObject]
#define	JOBIncPenatly(j, p)	\
	do { \
		NSMutableDictionary	*jDict = [self jobSpecificationForJobId:j]; \
		[jDict setObject:[NSNumber numberWithDouble:(p) + \
			[[jDict objectForKey:@"timePenalty"] doubleValue]] \
			forKey:@"timePenalty"]; \
	} while(0)

#define	ZillionOutcomesKey	@"outcomes"

- (NSMutableDictionary *)jobOutcomesForJobId:(NSString *)jobId
{
	NSMutableDictionary	*jobOutcomes = [outcomes objectForKey:jobId];

	if (!jobOutcomes) {
		jobOutcomes = [NSMutableDictionary dictionary];
		[outcomes setObject:jobOutcomes forKey:jobId];
		[jobOutcomes setObject:[NSMutableArray array] forKey:@"finishedSeeds"];
		[jobOutcomes setObject:[NSMutableDictionary dictionary] forKey:ZillionOutcomesKey];
	}
	return jobOutcomes;
}

- (void)submitOutcome:(NSData *)outcome ofJob:(NSString *)jobId forSeedId:(NSData *)seedId
	storeHints:(NSDictionary *)someStoreHints
{
	if (_respawnStatus) {
		id <ZillionServerProtocol>	proxyServer = PROXYSERVER_FOR_JOBID(jobId);
		[proxyServer submitOutcome:outcome ofJob:jobId forSeedId:seedId storeHints:someStoreHints];
	} else {
		id <ZillionJobProtocol>	job = [self jobInstance:jobId];

		[[job class] saveOutcome:outcome forSeedId:seedId storeHints:someStoreHints
			usingServer:[self masterServer]];
	}
}
/*
	we'll store outcomes if we are asked to do so
 */
- (void)setOutcome:(bycopy NSData *)outcome forSeedId:(bycopy NSData *)seedId
	andJobId:(bycopy NSString *)jobId
{
	NSMutableDictionary	*jobOutcomes = [self jobOutcomesForJobId:jobId];
	[[jobOutcomes objectForKey:ZillionOutcomesKey] setObject:outcome forKey:seedId];
}

- (NSMutableDictionary *)outcomeDictionaryForJobId:(NSString *)jobId
{
	return [[self jobOutcomesForJobId:jobId] objectForKey:ZillionOutcomesKey];
}

- (void)jobUnitDidEndForJobId:(NSString *)jobId seedId:(NSData *)seedId
	timePenalty:(NSTimeInterval)penalty
{
	NSMutableDictionary	*jobOutcomes = [self jobOutcomesForJobId:jobId];
	NSMutableSet		*seeds = [dispatchedSeeds objectForKey:jobId];

NSLog(@"about to Register finished seed %@", seedId);
	[[jobOutcomes objectForKey:@"finishedSeeds"] addObject:seedId];
	JOBIncPenatly(jobId, penalty);

	if (![seeds containsObject:seedId]) {
		[NSException raise:NSInternalInconsistencyException
			format:@"Finished Seed was not registered [%@]", seedId];
	} else {
		[seeds removeObject:seedId];
	}

	// are we ready to seal the job?
	if (![seeds count] && JobSeedStatus(jobId))
		SEALJOBWITHID(jobId);
	RESCHEDULE();
}

- (void)registerSeed:(NSData *)seedForJob forJob:(NSString *)jobId
{
	NSMutableSet	*seeds = [dispatchedSeeds objectForKey:jobId];

	if (!seeds) {
		seeds = [NSMutableSet set];
		[dispatchedSeeds setObject:seeds forKey:jobId];
	}
	[seeds addObject:[[seedForJob copy] autorelease]];
}


#if 0
- (double)cumulatedPenalty
{
	NSEnumerator	*jobEnum = [outcomes keyEnumerator];
	NSDictionary	*job;
	double	penalty = 0;

	while (job = [jobEnum nextObject]) {	
		penalty += JOBPenalty(job);
	}
	return penalty;
}
#endif

// this is a biggest residual seat distribution scheme where
//	dividend is the count of seats
//	relative is an array of (relative) votes
//	arrayLength is the number of parties and
//	result holds the number of seats after the distribution of seats

typedef struct {
	int		index;
	double	residual;
} Residual;
int cmpResidual(Residual *r1, Residual *r2) {
	return r1->residual == r2->residual? 0
		: (r1->residual < r2->residual? -1: 1);
}
typedef int (*QSortComparator)(const void*, const void*);

static void	scaleSetTo(double *relative, unsigned arrayLength, unsigned dividend, unsigned *result)
{	unsigned	i, cSum, cRest;
	double		rSum;
	Residual	residuals[arrayLength];

	for (i = 0, rSum = 0; i < arrayLength; i++) rSum += relative[i], result[i] = 0;
	if (!rSum) return;

	for (i = 0; i < arrayLength; i++)
		result[i] = (int)(relative[i] / rSum * arrayLength);

	do {
		for (i = 0; i < arrayLength; i++) {
			residuals[i] = ((Residual){i, result[i] / rSum * arrayLength - result[i]});
		}
		qsort(residuals, arrayLength, sizeof(Residual), (QSortComparator)cmpResidual);
		for (i = cSum = 0; i < arrayLength; i++) cSum += result[i];
		for (i = 0, cRest = dividend - cSum; i < arrayLength && cRest; i++, cRest--)
			result[residuals[i].index]++;
	} while (cRest);
}

// <i> reprioritize when a new job is started or an old one ends

#define	CLIENT_MAX_LOAD(c)	[[[c objectForKey:@"attributes"] objectForKey:@"load"] intValue]

- (void)scheduleJobs:sender
{
	NSEnumerator	*clientEnum;
	NSString		*clientId;
	NSMutableArray	*freeClients = [NSMutableArray array];

	NSLog(@"Scheduling jobs...");
	// clear flag
	_reschedulingIsMarked = NO;

	// determine which clients to run on
	for (clientEnum = [clients keyEnumerator]; (clientId = [clientEnum nextObject]); ) {
		NSDictionary	*clientSpec = [clients objectForKey:clientId];
		id <ZillionClientProtocol>	client = [clientSpec objectForKey:@"object"];
		int		i, count;
		int		currentLoad = [client currentLoad];
		for (i = 0, count = CLIENT_MAX_LOAD(clientSpec) - currentLoad ; i < count; i++)
			[freeClients addObject:client];
	}
NSLog(@"Free client count %d", [freeClients count]);
	if (![freeClients count]) return;

	NS_DURING
	// determine which jobs to run
	//	the inverse relative time penatly is proportional to the share in clients for each job
	{
		NSArray			*jobIds = [self jobIds];
		int				jobCount = [jobIds count];
		double			penalties[jobCount], penalty;
		NSString		*jobId;
		int				i, j, k, count, slots[jobCount];
		id <ZillionServerProtocol>	proxyServer;

		for (i = 0, count = [jobIds count]; i < count; i++) {
			jobId = [jobIds objectAtIndex:i];
			penalty = JOBPenalty(jobId);
			//<i> rather than setting to 1, set to priority of the job in the instance
			penalties[i] = penalty? 1/penalty: 1;
		}
		scaleSetTo(penalties, jobCount, [freeClients count], slots);

		for (i = k = 0, count = [jobIds count]; i < count; i++) {	
			jobId = [jobIds objectAtIndex:i];
			for (j = 0; j < slots[i]; j++) {
				NSData	*seedForJob;
				proxyServer = PROXYSERVER_FOR_JOBID(jobId);
				seedForJob = [proxyServer nextSeedIdForJobId:jobId];
				if (!seedForJob) {
					NSLog(@"Exhausted all seed for job %@", jobId);
					// we cannot now seal the job since we have to wait for all jobs to be finished
					// we therefore have to track which seeds were submitted and recognize
					// submission of results
					// we flag that we can now seal the job if all remaining seeds are done
					SetJobSeedsExhausted(jobId);
					break;
				}
				[self registerSeed:seedForJob forJob:jobId];
				NSLog(@"Dispatching unit for job %@", jobId);
				[[freeClients objectAtIndex:k++] dispatchUnitFromSeed:seedForJob forJobId:jobId];
			}
		}
	}
	NS_HANDLER
		NSLog(@"Exception raised during job dispatch with reason: %@", [localException reason]);
	NS_ENDHANDLER
	NSLog(@"did schedule...");
}

@end
